/*
 * Copyright (C) 2023-06-20  Julien BRUGUIER
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <string>
#include <sstream>

#line 135 "../plugins/tribe/core/tribe.svm_plugin"
#include <vector>
#include <map>
#line 26 "src/plugin.cpp"

#include <src/plugin.h>

extern "C"
{

void plugin_configure(void *plugin)
{
	::svm_plugin_configure(plugin,
		"PLUGIN tribe \n"
		"DEFINE \n"
		"	INTERRUPTION tribe.wrong_source \n"
		"	INTERRUPTION tribe.wrong_destination \n"
		"	INTERRUPTION tribe.interrupted \n"
		"	STRUCT tribe.identity \n"
		"	STRUCT tribe.message \n"
		"	TYPE tribe.process \n"
		"	WAITING INSTRUCTION tribe.process -> tribe.process \n"
		"	WAITING INSTRUCTION tribe.send STR:destination_name PTR:values \n"
		"	WAITING INSTRUCTION tribe.recv MUTABLE STR:source_name PTR:emplace ? -> PTR:values \n"
		"	SCHEDULER tribe.scheduler \n"
		,SVM_API_SIGNATURE,SVM_VERSION);
}

#ifndef ARGV_VALUE
#define ARGV_VALUE(index,type) ::svm_value_##type##_get(svm,::svm_parameter_value_get(svm,argv[(index)]))
#endif
#ifndef ARGV_PLUGIN
#define ARGV_PLUGIN(index,plugin,name) reinterpret_cast<type_##name*>(::svm_value_plugin_get_internal(svm,::svm_parameter_value_get(svm,argv[(index)])))
#endif
#ifndef ARGV_MARKER
#define ARGV_MARKER(index) std::string(::svm_parameter_marker_get(svm,argv[(index)]).string)
#endif
#ifndef ARGV_KEYWORD
#define ARGV_KEYWORD(index) std::string(::svm_parameter_keyword_get(svm,argv[(index)]).string)
#endif
#ifndef ARGV_STRUCT
#define ARGV_STRUCT(index,plugin,name) reinterpret_cast<struct_##name*>(::svm_structure_get_internal(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),::svm_parameter_structure_get(svm,argv[(index)])))
#endif
#ifndef ARGV_VARIABLE
#define ARGV_VARIABLE(index) ::svm_parameter_variable_get(svm,argv[(index)]);
#endif
#ifndef NEW_VALUE
#define NEW_VALUE(type,value) ::svm_value_##type##_new(svm,(value))
#endif
#ifndef NEW_PLUGIN
#define NEW_PLUGIN(plugin,name,value) ::svm_value_plugin_new(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),(value))
#endif
#ifndef NEW_STRUCT
#define NEW_STRUCT(plugin,name,value) ::svm_structure_new(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),(value))
#endif
#ifndef NEW_STRING
#define NEW_STRING(raw_string) ::svm_string_new(svm,raw_string.c_str(),raw_string.size())
#endif
#ifndef RAW_STRING
#define RAW_STRING(svm_string) std::string(svm_string.string,svm_string.size)
#endif
#ifndef NEW_BOOLEAN
#define NEW_BOOLEAN(boolean) ((boolean)?TRUE:FALSE)
#endif
#ifndef RAW_BOOLEAN
#define RAW_BOOLEAN(boolean) ((boolean)==TRUE)
#endif
#ifndef NEW_NULL_VALUE
#define NEW_NULL_VALUE(type) ::svm_value_##type##_new_null(svm)
#endif
#ifndef NEW_NULL_PLUGIN
#define NEW_NULL_PLUGIN(plugin,name) ::svm_value_plugin_new_null(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name))
#endif
#ifndef NEW_NULL_STRUCT
#define NEW_NULL_STRUCT(plugin,name) ::svm_structure_new_null(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name))
#endif
#ifndef ERROR_INTERNAL
#define ERROR_INTERNAL(irq,message) ::svm_processor_current_raise_error_internal__raw(svm,irq,message)
#endif
#ifndef ERROR_EXTERNAL
#define ERROR_EXTERNAL(plugin,name,message) ::svm_processor_current_raise_error_external__raw(svm,::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name),message)
#endif
#ifndef CONST_PEP
#define CONST_PEP(plugin,name) ::svm_value_pluginentrypoint_new__raw(svm,#plugin,#name)
#endif
#ifndef CURRENT
#define CURRENT(object) ::svm_##object##_get_current(svm)
#endif
#ifndef RETURN
#define RETURN return nullptr
#endif
#ifndef VARIABLE_GLOBAL
#define VARIABLE_GLOBAL(variable) ::svm_variable_scope_set_global(svm,(variable))
#endif
#ifndef VARIABLE_LOCAL
#define VARIABLE_LOCAL(variable) ::svm_variable_scope_set_local(svm,(variable))
#endif
#ifndef VARIABLE_DELETE
#define VARIABLE_DELETE(variable) ::svm_variable_delete(svm,(variable))
#endif

}

#line 141 "../plugins/tribe/core/tribe.svm_plugin"
	SVM_Event_Queue _events;
	std::map<std::string,SVM_Event_Queue_Address> _addresses;
	SVM_Lock _addresses_lock;
#line 130 "src/plugin.cpp"

extern "C"
{

/* INTERRUPTION tribe.wrong_source */


/* INTERRUPTION tribe.wrong_destination */


/* INTERRUPTION tribe.interrupted */


/* STRUCT tribe.identity */

struct struct_identity
{
#line 423 "../plugins/tribe/core/tribe.svm_plugin"
	explicit struct_identity(const std::string& name)
	:_name(name) {}
	std::string _name;
#line 152 "src/plugin.cpp"
};

void struct_identity_delete(const void *svm, void *handler)
{
	struct_identity * const object = reinterpret_cast<struct_identity*>(handler);
	{
#line 428 "../plugins/tribe/core/tribe.svm_plugin"

#line 161 "src/plugin.cpp"
	}
	delete object;
}


/* STRUCT tribe.message */

struct struct_message
{
#line 432 "../plugins/tribe/core/tribe.svm_plugin"
	std::vector<SVM_Value> _values;
#line 173 "src/plugin.cpp"
};

void struct_message_delete(const void *svm, void *handler)
{
	struct_message * const object = reinterpret_cast<struct_message*>(handler);
	{
#line 436 "../plugins/tribe/core/tribe.svm_plugin"
	for(auto& v:object->_values)
	{
		VARIABLE_LOCAL(v);
	}
#line 185 "src/plugin.cpp"
	}
	delete object;
}

}

/* code */

#line 445 "../plugins/tribe/core/tribe.svm_plugin"
	struct Tools
	{
		static void add_address(const void *svm, const std::string& process_name)
		{
			SVM_LockGuard_Write write = ::svm_lock_writeguard_new(svm,_addresses_lock,FALSE);
			SVM_Event_Queue_Address address = nullptr;
			auto it=_addresses.find(process_name);
			if(it!=_addresses.end())
			{
				address = it->second;
			}
			else
			{
				SVM_Structure identity = ::svm_structure_new(svm,CONST_PEP(tribe,identity),new struct_identity(process_name));
				address = ::svm_event_address_new_struct(svm,identity);
				::svm_variable_scope_set_shared(svm,address);
				::svm_event_queue_join(svm,_events,address);
			}
			VARIABLE_GLOBAL(address);
			_addresses.insert(std::make_pair(process_name,address));
		}
		static bool remove_address(const void *svm, const std::string& process_name)
		{
			SVM_LockGuard_Write write = ::svm_lock_writeguard_new(svm,_addresses_lock,FALSE);
			auto it = _addresses.find(process_name);
			if(it==_addresses.end()) return false;
			SVM_Event_Queue_Address address = it->second;
			VARIABLE_LOCAL(address);
			if(::svm_variable_scope_is_local(svm,address))
			{
				::svm_event_queue_leave(svm,_events,address);
				_addresses.erase(it);
			}
			return true;
		}
	};
#line 231 "src/plugin.cpp"

extern "C"
{

/* TYPE tribe.process */

struct type_process
{
#line 485 "../plugins/tribe/core/tribe.svm_plugin"
	explicit type_process(const std::string& name)
	:_name(name) { }

	operator std::string () const
	{
		return _name;
	}
	std::string _name;
#line 249 "src/plugin.cpp"
};

void type_process_delete(const void *svm, void *handler)
{
	type_process * const object = reinterpret_cast<type_process*>(handler);
	{
#line 496 "../plugins/tribe/core/tribe.svm_plugin"
	Tools::remove_address(svm,object->_name);
#line 258 "src/plugin.cpp"
	}
	delete object;
}

SVM_String type_process_print(const void *svm, const void *handler)
{
	const type_process *object = reinterpret_cast<const type_process*>(handler);
	std::string string = static_cast<std::string>(*object);
	{
#line 499 "../plugins/tribe/core/tribe.svm_plugin"

#line 270 "src/plugin.cpp"
	}
	return ::svm_string_new(svm,string.c_str(),string.size());
}


/* WAITING INSTRUCTION tribe.process -> tribe.process */

SVM_Value instruction_process(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 510 "../plugins/tribe/core/tribe.svm_plugin"
	SVM_String n = ::svm_process_get_name(svm,CURRENT(process));
	auto rn = RAW_STRING(n);
	auto a = new type_process(rn);
	Tools::add_address(svm,rn);
	return NEW_PLUGIN(tribe,process,a);
#line 286 "src/plugin.cpp"
}


/* WAITING INSTRUCTION tribe.send STR:destination_name PTR:values */

SVM_Value instruction_send(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 525 "../plugins/tribe/core/tribe.svm_plugin"
	auto destination = ARGV_VALUE(0,string);
	auto destination_raw = RAW_STRING(destination);
	auto pointer = ::svm_parameter_value_get(svm,argv[1]);
	auto message_raw = new struct_message();
	SVM_Structure message = ::svm_structure_new(svm,CONST_PEP(tribe,message),message_raw);
	SVM_Address address = ::svm_value_pointer_get_address(svm,pointer);
	SVM_Size size = ::svm_value_pointer_get_size(svm,pointer);
	for(SVM_Address a=address ; a<(address+size) ; ++a)
	{
		SVM_Value value = ::svm_memory_read_address(svm,CURRENT(kernel),a);
		VARIABLE_GLOBAL(value);
		message_raw->_values.push_back(value);
		::svm_memory_write_address(svm,CURRENT(kernel),a,::svm_value_automatic_new_null(svm));
	}
	SVM_LockGuard_Read read = ::svm_lock_readguard_new(svm,_addresses_lock,FALSE);
	auto source = ::svm_process_get_name(svm,CURRENT(process));
	auto self = _addresses.find(RAW_STRING(source));
	if(self==_addresses.end())
	{
		ERROR_EXTERNAL(tribe,wrong_source,"Source is not a process within the tribe");
	}
	auto it = _addresses.find(destination_raw);
	if(it==_addresses.end())
	{
		ERROR_EXTERNAL(tribe,wrong_destination,"Destination is not a process within the tribe");
	}
	VARIABLE_DELETE(read);
	::svm_event_queue_push(svm,_events,it->second,self->second,message);
#line 323 "src/plugin.cpp"
	return nullptr;
}


/* WAITING INSTRUCTION tribe.recv MUTABLE STR:source_name PTR:emplace ? -> PTR:values */

SVM_Value instruction_recv(const void *svm, SVM_Size argc, SVM_Parameter argv[])
{
#line 569 "../plugins/tribe/core/tribe.svm_plugin"
	auto source = ::svm_parameter_value_get(svm,argv[0]);
	SVM_Value_Pointer pointer = nullptr;
	if(argc>1)
	{
		pointer = ::svm_parameter_value_get(svm,argv[1]);
	}
	SVM_LockGuard_Read read = ::svm_lock_readguard_new(svm,_addresses_lock,FALSE);
	auto destination = ::svm_process_get_name(svm,CURRENT(process));
	auto self = _addresses.find(RAW_STRING(destination));
	if(self==_addresses.end())
	{
		ERROR_EXTERNAL(tribe,wrong_destination,"Destination is not a process within the tribe");
	}
	SVM_Event_Queue_Address destination_address = self->second;
	VARIABLE_DELETE(read);
	SVM_Event_Queue_Address s;
	SVM_Structure message;
	::svm_process_interruptionnotification_enable(svm,CURRENT(process));
	SVM_Boolean result = ::svm_event_queue_pull(svm,_events,destination_address,&s,&message,TRUE);
	::svm_process_interruptionnotification_disable(svm,CURRENT(process));
	if(not result)
	{
		ERROR_EXTERNAL(tribe,interrupted,"Receive interrupted");
	}
	auto source_raw = reinterpret_cast<struct_identity*>(::svm_structure_get_internal(svm,CONST_PEP(tribe,identity),::svm_event_address_get_struct(svm,s)));
	auto message_raw = reinterpret_cast<struct_message*>(::svm_structure_get_internal(svm,CONST_PEP(tribe,message),message));
	if(not pointer)
	{
		SVM_Memory_Zone zone = ::svm_memory_zone_new(svm);
		::svm_memory_zone_append_internal__raw(svm,zone,AUTOMATIC,message_raw->_values.size());
		pointer = ::svm_memory_allocate(svm,CURRENT(kernel),zone);
	}
	else
	{
		SVM_Address address = ::svm_value_pointer_get_address(svm,pointer);
		SVM_Size size = ::svm_value_pointer_get_size(svm,pointer);
		if(message_raw->_values.size()!=size)
		{
			ERROR_INTERNAL(MEMORY,"Provided pointer has invalid size");
		}
		auto it = message_raw->_values.cbegin();
		for(SVM_Address a=address ; a<(address+size) ; ++a,++it)
		{
			if(not ::svm_memory_address_is_writable(svm,CURRENT(kernel),a,*it))
			{
				ERROR_INTERNAL(MEMORY,"Provided value has invalid type");
			}
		}
	}
	SVM_Address address2 = ::svm_value_pointer_get_address(svm,pointer);
	SVM_Size size2 = ::svm_value_pointer_get_size(svm,pointer);
	auto it2 = message_raw->_values.cbegin();
	for(SVM_Address a=address2 ; a<(address2+size2) ; ++a,++it2)
	{
		::svm_value_state_set_movable(svm,*it2);
		::svm_memory_write_address(svm,CURRENT(kernel),a,*it2);
	}
	::svm_value_string_set(svm,source,NEW_STRING(source_raw->_name));
	return pointer;
#line 392 "src/plugin.cpp"
}


/* SCHEDULER tribe.scheduler */

struct scheduler_scheduler
{
#line 646 "../plugins/tribe/core/tribe.svm_plugin"
	std::vector<SVM_Process> _processes;
#line 402 "src/plugin.cpp"
};

void* scheduler_scheduler_create(const void *svm)
{
	scheduler_scheduler *object = new scheduler_scheduler();
	{
#line 649 "../plugins/tribe/core/tribe.svm_plugin"

#line 411 "src/plugin.cpp"
	}
	return object;
}

void scheduler_scheduler_delete(const void *svm, void *handler)
{
	scheduler_scheduler * const object = reinterpret_cast<scheduler_scheduler*>(handler);
	{
#line 651 "../plugins/tribe/core/tribe.svm_plugin"
	for(auto& p:object->_processes)
	{
		VARIABLE_LOCAL(p);
	}
#line 425 "src/plugin.cpp"
	}
	delete object;
}

unsigned long int scheduler_scheduler_schedule(const void *svm, void *handler, const SVM_Process process, const SVM_Process_State state)
{
	scheduler_scheduler *object = reinterpret_cast<scheduler_scheduler*>(handler);
	{
#line 658 "../plugins/tribe/core/tribe.svm_plugin"
	if((state==SUSPENDED) or (state==CONTINUE))
	{
		::svm_process_run__raw(svm,process,0);
	}
	return 0;
#line 440 "src/plugin.cpp"
	}
}

SVM_Boolean scheduler_scheduler_attach(const void *svm, void *handler, const SVM_Process process, unsigned long int parameter)
{
	scheduler_scheduler *object = reinterpret_cast<scheduler_scheduler*>(handler);
	{
#line 666 "../plugins/tribe/core/tribe.svm_plugin"
	auto process_name = RAW_STRING(::svm_process_get_name(svm,process));
	Tools::add_address(svm,process_name);
	VARIABLE_GLOBAL(process);
	object->_processes.push_back(process);
	return TRUE;
#line 454 "src/plugin.cpp"
	}
	return TRUE;
}

SVM_Boolean scheduler_scheduler_detach(const void *svm, void *handler, const SVM_Process process, unsigned long int parameter)
{
	scheduler_scheduler *object = reinterpret_cast<scheduler_scheduler*>(handler);
	{
#line 674 "../plugins/tribe/core/tribe.svm_plugin"
	auto process_name = RAW_STRING(::svm_process_get_name(svm,process));
	if(not Tools::remove_address(svm,process_name))
	{
		return FALSE;
	}
	for(auto p=object->_processes.begin() ; p!=object->_processes.end() ; ++p)
	{
		if(*p==process)
		{
			VARIABLE_LOCAL(*p);
			object->_processes.erase(p);
			break;
		}
	}
	return TRUE;
#line 479 "src/plugin.cpp"
	}
	return TRUE;
}

SVM_String scheduler_scheduler_print(const void *svm, const void *handler)
{
	const scheduler_scheduler *object = reinterpret_cast<const scheduler_scheduler*>(handler);
	{
#line 692 "../plugins/tribe/core/tribe.svm_plugin"
	std::ostringstream oss;
	for(auto& p:object->_processes)
	{
		oss << RAW_STRING(::svm_process_print(svm,p)) << std::endl;
	}
	return NEW_STRING(oss.str());
#line 495 "src/plugin.cpp"
	}
}


/* Generic handling functions */

void plugin_initialisation(const void *svm)
{
#line 148 "../plugins/tribe/core/tribe.svm_plugin"
	_events = ::svm_event_queue_new(svm);
	VARIABLE_GLOBAL(_events);
	_addresses_lock = ::svm_lock_new(svm);
	VARIABLE_GLOBAL(_addresses_lock);
#line 509 "src/plugin.cpp"
}

void plugin_finalisation(const void *svm)
{
#line 156 "../plugins/tribe/core/tribe.svm_plugin"
	VARIABLE_LOCAL(_events);
	for(auto& a: _addresses)
	{
		VARIABLE_LOCAL(a.second);
	}
	VARIABLE_LOCAL(_addresses_lock);
#line 521 "src/plugin.cpp"
}

}
